#!/usr/bin/python2

"""
A simple yet powerfull one file WikiWiki clone

Copyright 2003 Graham King

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

import cgi
import re
import os
import os.path
import glob
import stat
import time
import cPickle
import smtplib
import email.Utils

##
## Load the configuration
##
from scribbleconfig import *


if USE_RCS:
        import rcslib

# The colors to display people's names as
nameColorHash = {}

# The filename with full path to use when saving the name colors 
nameColorFilename = repository + "nameColor.pickle"

# The filename with full path to save email update info to
emailUpdateFilename = repository + "emailUpdate.pickle"

# E-mail frequency statics. The only interesting one to change is FREQUENCY_DEFAULT

FREQUENCY = "frequency"
FREQUENCY_DAILY = "daily"
FREQUENCY_WEEKLY = "weekly"
FREQUENCY_DEFAULT = FREQUENCY_DAILY

##
## Date manipulation methods
##

daysInMonth = {1: 31, 2: 28, 3: 31, 4: 30, 5: 31, 6: 30,
                7: 31, 8: 31, 9: 30, 10: 31, 11: 30, 12: 31 }

#
# isLeapYear: Returns true if the year number given is
#  a leap year

def isLeapYear(year):

    true = 1
    false = 0

    if ( year % 4 != 0 ):
        return false
    elif ( year % 400 == 0 ):
        return true
    elif ( year % 100 == 0 ):
        return false
    else:
        return true


#
# monthsToDays: Converts the given number of months
# to a number of days since January.
#
# Note that this month is part of a date,
# so we do not count the current month (because
# we are only part way through it). This means January
# will return 0, February will return the number of days
# in January, etc.

def monthsToDays(months, isInLeapYear):

    totalDays = 0

    for currentMonth in range(1,months):

        totalDays += daysInMonth[currentMonth]

        # If we included february and its a leap year add the extra day

        if ( months >= 3 and isInLeapYear ):
            totalDays += 1

    return totalDays

#
# yearsToDays: Converts years to days since 1900
#

def yearsToDays(years):

    totalDays = 0
  
    for currentYear in range(1900, years):

        if ( isLeapYear(currentYear) ):
            totalDays += 366
        else:
            totalDays += 365

    return totalDays


#
# daysSinceBase: Return the number of days elapsed since 1st Jan 1900
# So 1st Jan 1900 will return 0, 2nd Jan 1900 will return 1, etc.
#

def daysSinceBase(days, months, years):

    totalDays = days - 1

    totalDays += monthsToDays(months, isLeapYear(years) )

    totalDays += yearsToDays(years)

    return totalDays


# 'date' should be a tuple output from time.localtime(xx)
def isToday(date):

    today = time.localtime( time.time() )

    thisYear = ( today[0] == date[0] )
    thisMonth = ( today[1] == date[1] )
    thisDay = ( today[2] == date[2] )

    return ( thisYear and thisMonth and thisDay )

def isYesterday(date):

    today = time.localtime( time.time() )

    dayCountDate = daysSinceBase( date[2], date[1], date[0] )
    dayCountToday = daysSinceBase( today[2], today[1], today[0] )

    return ( dayCountDate == dayCountToday - 1 )

def isThisWeek(date):

    today = time.localtime( time.time() )

    dayCountDate = daysSinceBase( date[2], date[1], date[0] )
    dayCountToday = daysSinceBase( today[2], today[1], today[0] )

    return ( dayCountDate >= dayCountToday - 7 )

##
## Filename and string manipulation
##

# Is the filename given one we should allow. 
# Prevents people entering things like '../sfsdf' or '/etc/blah' as the filename.
def isValidFilename(filename):

        allowedCharacters = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890-_.~"

        # Prevent the attempt to go up a directory.
        # We need this on top of the usual check as we have to allow '.' in filenames

        if filename.find("..") != -1:
                return 0


        # Check all the characters are allowed
        for ch in filename:
            if not ch in allowedCharacters:
                    return 0

        return 1

# Turns a filename into one we should allow. 
# Similar to isValidFilename
def makeValidFilename(filename):

        allowedCharacters = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890-_.~"

        result = ""

        # Prevent the attempt to go up a directory.

        filename = filename.replace("..", "")

        for ch in filename:
            if ch in allowedCharacters:
                result += ch

        return result 

"""
 Is the 'string' present in the file called 'filename'. Case insensitive.
 """
def isStringPresent(string, filename):

    file = open(filename, "rt")
    wholefile = file.read().lower()
    file.close()
    
    return ( wholefile.find(term.lower()) != -1 )
    

"""
 The color to display a particular name as
"""

def getDisplayColor(name):

    if len(nameColorHash) == 0:

        # Ensure pickle if exists
        if not os.path.exists(nameColorFilename):
            save(nameColorFilename, nameColorHash)

        nameColorHash.update( load(nameColorFilename) )

    if not nameColorHash.has_key(name):
        nameColorHash[name] = DEFAULT_NAME_COLOR
        save(nameColorFilename, nameColorHash)

    return nameColorHash[name]

"""
Load (unpickle) from 'filename', which must include full path,
and return the data loaded
"""
def load(filename):

    file = open( filename, "rb")
    data = cPickle.load(file)
    file.close() 

    return data

"""
 Save (pickle) 'data' to 'filename'. 'filename' must include full path
"""
def save(filename, data):

    file = open(filename, "wb")
    cPickle.dump(data, file, 1)
    file.close()

##
## Wiki methods
##

"""
Transforms a wiki string into an HTML string. 
In practice this replaces WikiWords with <a href link, and *word* with <b>word</b>, and that sort of stuff.
"""
def toHTML(string):

    # Links
    string = re.sub("[\s+|^]([A-Z]+[a-z]+[A-Z]+\w*)", linkReplace, string)

    # Bold
    string = re.sub("(\*[\w| ]*\*\s+?)", boldReplace, string)

    # Italic
    string = re.sub("(\_[\w| ]*\_\s+?)", italicReplace, string)

    # Header
    header = re.compile("^(---\+[\w| ]+)", re.MULTILINE)
    string = re.sub(header, headerReplace, string)

    # Sub header
    subheader = re.compile("^(---\+\+[\w| ]+)", re.MULTILINE)
    string = re.sub(subheader, subheaderReplace, string)

    # IMG links
    string = re.sub("[\s+|^](image:\S*)", imageLinkReplace, string)

    # Outside links
    string = re.sub("(\s{1})(http://\S*)", outsideLinkReplace, string)

    # Horizontal lines
    dashes = re.compile("(^----\\s+)", re.MULTILINE)
    string = re.sub(dashes, lineReplace, string)

    # Email addresses
    string = re.sub("([\w|\.]+@\w+\.[\w|\.]+)", emailReplace, string)

    # People's name
    nameInSquareBrackets = re.compile("^\[(\w*)\]", re.MULTILINE)
    string = re.sub(nameInSquareBrackets, nameColor, string)

    # Special keys
    string = re.sub("##([A-Z_]*)##", specialKeyCaller, string)

    #
    # Line by line processing
    #
    result = []
    inTable = 0
    tableString = ""
    for line in string.split('\n'):

        foundPattern = re.search("^\s*\|.*\|.*\|\s$", line)

        if foundPattern:

            if not inTable:
                tableString = "<table border=1>"
                inTable = 1

            line = re.sub("^\s*\|", "<tr><td>", line)      # Start of line
            line = re.sub("\|\s*$", "</td></tr>", line)    # End of line
            line = re.sub("\|", "</td><td>", line)      # Middle of line

            tableString += line

        else:
            if inTable:
                tableString += "</table>"
                inTable = 0
                result.append(tableString)

            result.append(line)

    # In case we had a table at the end of file
    if inTable:
        string.append("</table>")
        inTable = 0

    string = '\n'.join(result) 

    # Line breaks
    breaks = re.compile("($)", re.MULTILINE)
    string = re.sub(breaks, breakReplace, string)

    return string


# In the methods below the +1 at the start gets rid of the initial
# character (the asterix or underline or whatever) and the -2
# at the end gets rid of the second char (asterix, underline, etc) 
# and of the whitespace or newline.
# Notice that we re-insert a whitespace after the closing tag
# to keep the word separated from others.

def headerReplace(matchObject):

    return "<h2>"+ matchObject.group(1)[+4:] +"</h2>"

def subheaderReplace(matchObject):

    return "<h3>"+ matchObject.group(1)[+5:] +"</h3>"

def italicReplace(matchObject):

    return "<i>"+ matchObject.group(1)[+1:-2] +"</i> "

def emailReplace(matchObject):

        return " <a href=\"mailto:"+ matchObject.group(1) +"\">"+ matchObject.group(1)+"</a>"


def boldReplace(matchObject):

    return "<b>"+ matchObject.group(1)[+1:-2] +"</b> "

def nameColor(matchObject):

    name = matchObject.group(1)
    color = getDisplayColor(name)
    return "[<font color=\""+ color +"\">"+ name +"</font>]"

def breakReplace(matchObject):

    return "<br>"

def linkReplace(matchObject):

    result = ''

    page = matchObject.group(1)

    filename = fullpath(page)

    action = "view"
    pagename = page

    # Pages not yet created have to be edited first
    if not os.path.exists(filename):
        action = "edit"
        pagename += "?"

    result += "<a href=\""+ scriptURL +"?action="+ action +"&page="+ page +"\">"
    result += " "+ pagename     # We preserve the first space
    result += "</a>"

    return result

def imageLinkReplace(matchObject):

    url = matchObject.group(1)[6:]

    return "<img src=\""+ url +"\">"

def outsideLinkReplace(matchObject):

    separator = matchObject.group(1)
    url = matchObject.group(2)

    return separator +"<a href=\""+ url +"\">"+ url +"</a>"

def lineReplace(matchObject):

    return "<hr>"

def specialKeyCaller(matchObject):

    specialKey = matchObject.group(1)
    method = specialKeyToMethodMap[specialKey]
    return method()

def view(filename, page):

    file = open( filename, "rt" )
    contents = file.readlines()
    file.close()

    result = ""

    result += "<h2>"
    result += "<a href=\""+ scriptURL +"?action=search&term="+ page +"\">"
    result += page +"</a>"
    result += "</h2>\n"

    result += toHTML( "".join(contents) )

    result += toolbar(page)

    return result

def toolbar(page):

    result = "<br><hr>\n"

    result += "<table cellspacing=5><tr>\n"

    if page != None and page != "": 
        result += "<td>\n"
        result += "<form action=\""+ scriptURL +"\">\n"
        result += "<input type=hidden name=action value=edit>\n"
        result += "<input type=hidden name=page value=\""+ page +"\">\n"
        result += "<input type=submit value=\"Edit\">\n"
        result += "</form>\n"
        result += "</td>\n"

    result += "<td><a href=\""+ scriptURL +"\">Home</a></td>\n"

    result += "<td>\n"
    result += "<form action=\""+ scriptURL +"\">\n"
    result += "<input type=hidden name=action value=search>\n"
    result += "<input type=text name=term>\n"
    result += "<input type=submit value=\"Search\">\n"
    result += "</form>\n"
    result += "</td>\n"

    result += "<td>\n"
    result += "<a href=\""+ scriptURL +"?action=whatsnew\">"
    result += "What's New</a>\n"
    result += "</td>\n"

    if page != None and USE_RCS:
            result += "<td>\n"
            result += "<a href=\""+ scriptURL +"?action=diff&page="+ page +"\">"
            result += "What's changed in this page</a>\n"
            result += "</td>\n"

    result += "</tr></table>\n"

    return result

def edit(filename, page):

    if os.path.exists(filename):

        file = open( filename, "rt" )
        contents = file.readlines()
        file.close()
    else:
        contents = []
    
    result = "<form action=\""+ scriptURL +"\" method=\"post\">\n"

    result += "<input type=hidden name=action value=save>\n"
    result += "<input type=hidden name=page value=\""+ page +"\">\n"

    result += "<textarea rows=25 cols=70 name=body>\n"
    result += "".join(contents)
    result += "</textarea>\n"
    result += "<br>\n"
    result += "<input type=submit value=\"Save\">\n"
    result += "</form>\n"
    
    result += "<form action=\""+ scriptURL +"\" method=\"post\" enctype=\"multipart/form-data\">\n"
    result += "<input type=hidden name=action value=saveimage>\n"
    result += "<input type=hidden name=page value=\""+ page +"\">\n"
    result += "<input type=file name=image>\n"
    result += "<input type=submit value=\"Upload image\"> <font size=-2>Upload image before typing text</font>\n"
    result += "</form>\n"

    result += edithelp()


    return result

def search(term):

    numResults = 0

    result = ""

    result += "<h2>Search results</h2>"
    result += "Search results for '"+ term +"':<br><br>\n"

    allfiles = glob.glob( repository + os.sep +"*" )
    for filename in allfiles:

        if filename.endswith(",v"): # Ignore version control files
            continue

        if isStringPresent(term, filename):
            numResults += 1 

            pagename = os.path.basename(filename)
            result += "<a href=\""+ scriptURL
            result += "?action=view&page="+ pagename +"\">"
            result += pagename
            result += "</a><br>\n"

    if numResults == 0:
        result += "No results found\n"
    else:
        result+= "<br>"+ str(numResults) +" result(s)\n"

    result += toolbar(None)

    return result



def edithelp():

    return """
    <small>
    Formatting rules:
    <ul>
        <li>WikiWord -> Link or new page
        <li>[name] -> Your name in color
        <li>*word* -> <b>bold</b>
        <li>_word_ -> <i>italic</i>
        <li>http://address.org -> Outside link
        <li>image:http://address.org/picture.jpg -> Image on remote server
        <li>bob@something.com -> mail to link
        <li>---+ word -> header (three dashes and a plus at start of line)
        <li>---++ word -> sub header (three dashes and two pluses at start of line)
        <li>---- -> Horizontal rule (four dashes at start of line followed by space)
        <li>|one|two| -> Table
    </ul>
    </small><br>
    """

def isImage(filename):

    return ( filename.endswith(".jpg") or \
             filename.endswith(".jpeg") or \
             filename.endswith(".gif") or \
             filename.endswith(".png") )

def isPageFile(filename):

    if filename.endswith(",v"): # Skip version control files
        return 0

    if filename.endswith(".pickle"): # Skip saved state
        return 0 

    if IGNORE_IMAGES and isImage(filename): # Skip images
        return 0 

    return 1

def pageList():

    return filter( isPageFile, os.listdir( repository ) )

"""
 Takes a list of pages and returns a dictionary where the key is the time that
  file changed and the value the page name. 
"""
def timePagesChanged(allfiles):

    changeTimeDict = {}

    for filename in allfiles:

        fullname = fullpath( filename )
        lastModified = time.localtime( os.stat(fullname)[stat.ST_MTIME] )

        changeTimeDict[ lastModified ] = filename

    return changeTimeDict

def whatsnew():

    result = "<h2>What's new</h2>\n"

    result += "<b>Most recent changes, newest first: </b><br>"
    result += "<br>\n"

    changeTimeDict = timePagesChanged( pageList() )

    dates = changeTimeDict.keys()
    dates.sort()
    dates.reverse()

    if len(dates) > MAX_WHATS_NEW:
        dates = dates[:MAX_WHATS_NEW]

    for date in dates:

        result += time.asctime(date) +": "
        page = changeTimeDict[ date ]

        if isImage(page):

            viewcommand = "viewimage"
        else:

            viewcommand = "view"
        
        result += "<a href=\""+ scriptURL +"?action="+ viewcommand +"&page="+ page +"\">"
        result += page +"</a>"

        if isToday(date):
            result += "&nbsp;<i>(Today)</i>"
        elif isYesterday(date):
            result += "&nbsp;<i>(Yesterday)</i>"
        elif isThisWeek(date):
            result += "&nbsp;<i>(This week)</i>"

        result += "<br>\n"

    result += toolbar(None)

    return result

"""
The HTML header, including the sytlesheet link
"""
def header(title):

    header = ""

    header += "<head>\n"
    header += "<title>Scribble - "+ title +"</title>"

    if STYLESHEET is not None and STYLESHEET != "":
        header += "<link rel=STYLESHEET TYPE=\"text/css\" HREF=\""+ STYLESHEET +"\">"

    header += "</head>\n"

    return header

"""
The HTML body start, including the top logo and opening the table cell
that contains the whole page
"""
def bodyStart():

    bodyHTML = "<body>\n"

    logoHTML = """
        <table id="topbar">
        <tr>
            <td align="right">
                <img src=\""""+ LOGO + """\" align="right">
            </td>
        </tr>
        </table>
        """

    tableStart = """
        <table cellpadding="10" width="100%">
            <tr>
                <td>
        """

    if LOGO is not None and LOGO != "":
        return bodyHTML + logoHTML + tableStart
    else:
        return bodyHTML + tableStart

 
"""
Close what bodyStart opened and the whole html tag
"""
def bodyEnd():

    tableEnd = "</td></tr></table>\n"
    bodyEnd = "</body>\n"

    return tableEnd + bodyEnd + "</html>"


"""
'back' is how many revisions back to compare. If 'page' is on revision 1.5 then
'back' of 0 compares 1.5 and 1.4. 'back' of 1 compares 1.4 and 1.3, etc.
"""
def rcsdiff(page, back=0):

    path = fullpath(page)
    
    rcs = rcslib.RCS()
    
    result = "<h2>Most recent changes within '"+ page +"'</h2>"
    result += "Lines starting with &gt; were ADDED<br>\n"
    result += "Lines starting with &lt; were REMOVED<br>\n"
    result += "<br>\n"

    try:
        revision = rcs.calculateVersionNumber(path, back)
        rcsoutput = rcs.diff( path, revision )

        # HTMLify

        rcsoutput = rcsoutput.replace("<", "&lt;")
        rcsoutput = rcsoutput.replace(">", "&gt;")

        breaks = re.compile("($)", re.MULTILINE)
        rcsoutput = re.sub(breaks, breakReplace, rcsoutput)

        result += rcsoutput

    except NameError:
        result += "No changes. New file.<br>\n"
    except OSError:
        result += "File not yet in version control, so cannot obtain changes.<br>\n"

    
    result += "<a href=\""+ scriptURL
    result += "?action=diff"
    result += "&page="+ page
    result += "&back="+ str(back+1) +"\">"
    result += "Previous changes</a>\n"

    result += "<br>\n"
    result += toolbar(page)

    return result

def saveimage(image, page):

    filename = image.filename

    if filename.find("/") != -1:
        filename = filename[ filename.rfind("/"): ]

    if filename.find("\\") != -1:
        filename = filename[ filename.rfind("\\"): ]

    filename = makeValidFilename( filename )

    if not ( filename.endswith(".jpg") or \
             filename.endswith(".jpeg") or \
             filename.endswith(".gif") or \
             filename.endswith(".png") ):

                 raise ValueError("Wrong file type. Must be jpeg, gif or png.")

    # Calculate the size

    alllines = image.file.readlines()
    size = 0
    for line in alllines:
        size += len(line)

    if size > MAX_IMAGE_SIZE:
        errorMessage = "Image too big."
        errorMessage += "Maximum size is: "+ str(MAX_IMAGE_SIZE) 
        errorMessage += " ("+ str( MAX_IMAGE_SIZE/1024 ) +" Kb)"
        raise ValueError(errorMessage)

    # Save the image
    out = open( fullpath(filename), "wb" )
    out.writelines( alllines )
    out.close()

    # Add a link to the page

    pagepath = fullpath(page)

    if USE_RCS:

            try:
                rcs = rcslib.RCS()
                rcs.checkout(pagepath, 1, "-f")

            except OSError:
                # This means file not yet in version control. Ignore
                None

    wikipage = open( pagepath, "rt" )
    contents = wikipage.readlines()
    wikipage.close()

    contents.append("\n<!-- The text below is the image you have uploaded. Move that to where you want the image to appear in the page -->\n")
    contents.append( \
        "image:"+ scriptURL +"?action=viewimage&page="+ filename +"\n")

    wikipage = open( fullpath(page), "wt" )
    wikipage.writelines( contents )
    wikipage.close()

    if USE_RCS:
        rcs = rcslib.RCS()
        rcs.checkin(pagepath)


"""
The full absolute path of 'file'
"""
def fullpath(file):

    if repository.endswith( os.sep ):
        return repository + file
    else:
        return repository + os.sep + file

def viewimage(filename):

    if filename.endswith(".jpg") or filename.endswith(".jpeg"):
        result = "Content-type: image/jpeg"
    elif filename.endswith(".gif"):
        result = "Content-type: image/gif"
    elif filename.endswith(".png"):
        result = "Content-type: image/png"
    else:
        raise ValueError("Invalid image type. Must be jpeg, gif or png.")

    result += "\n\n"

    file = open(filename, "rb")
    result += file.read()
    file.close()

    return result

def setColors(form):

    for key in form.keys():

        if key == "action" or key == "page":
            continue

        name = key
        color = form[key].value

        nameColorHash[name] = color

    save(nameColorFilename, nameColorHash)

def colorInputTable():

    html = ""

    html += "<form action="+ scriptURL +">"
    html += "<input type=hidden name=action value=colors>"
    html += "<input type=hidden name=page value="+ page +">"
    html += "<table>"
    for name in nameColorHash.keys():
        html += "<tr><td>"+ name +"</td>"
        html += "<td>"
        html += "<input type=text name="+ name +" value="+ nameColorHash[name] +">"
        html += "</td></tr>"

    html += "<tr><td colspan=2>&nbsp;</td></tr>\n"
    html += "<tr><td colspan=2 align=center>"
    html += "<input type=submit value=Save></td></tr>\n"
    html += "</table>"
    html += "</form>"

    return html

def emailUpdateForm():

    emailUpdateDict = {}
    if os.path.exists(emailUpdateFilename):
        emailUpdateDict = load(emailUpdateFilename) 

    html = ""

    html += "<form action="+ scriptURL +">"
    html += "<input type=hidden name=page value="+ page +">"
    html += "<table>"

    if emailUpdatesPageList == []:  # Find
        html += "<input type=hidden name=action value=findEmailUpdates>"
        html += "<tr><td>Email address:</td>"
        html += "<td><input type=text name=email></td></tr>"
        html += "<tr><td colspan=2><input type=submit value=Find></td></tr>"
    else:   # Edit and save

        # Order page list by most recent activity at top

        timeChangeDict = timePagesChanged(emailUpdatesPageList)
        dates = timeChangeDict.keys() 
        dates.sort()
        dates.reverse()
        orderedPages = []
        for date in dates:
            orderedPages.append( timeChangeDict[date] )

        html += "<tr><td>Email address:</td>"
        html += "<td>"+ email +"</td></tr>"

        pageChoiceList = []
        if emailUpdateDict.has_key(email):
            pageChoiceList = emailUpdateDict[email]

        html += "<input type=hidden name=action value=saveEmailUpdates>"
        html += "<input type=hidden name=email value="+ email +">"
        for pageName in orderedPages:
            html += "<tr>"
            html += "<td>"+ pageName +"</td>"
            html += "<td><input type=checkbox name="+ pageName 
            if pageName in pageChoiceList:
                html += " checked"
            html += "></td>"
            html += "</tr>"

        frequency = FREQUENCY_DEFAULT
        if emailUpdateDict.has_key(email + FREQUENCY):
            frequency = emailUpdateDict[email + FREQUENCY]

        html += "<tr><td colspan=2>&nbsp;</td></tr>"
        html += "<tr><td colspan=2>"
        html += "<table><tr>"
        html += "<td>Send updates:</td>"
        html += "<td>Daily ? <input type=radio name="+ FREQUENCY +" value="+ FREQUENCY_DAILY
        if frequency == FREQUENCY_DAILY:
            html += " checked"
        html += "></td>"
        html += "<td>Weekly ? <input type=radio name="+ FREQUENCY +" value="+ FREQUENCY_WEEKLY
        if frequency == FREQUENCY_WEEKLY:
            html+= " checked"
        html += "></td>"
        html += "</tr></table>"
        html += "</td></tr>"
        html += "<tr><td colspan=2><input type=submit value=Save></td></tr>"

    html += "</table>"
    html += "<br>"

    # List current users

    html += "<table>" 
    html += "<tr><td><b>User</b><td><b>Pages</b></td><td><b>Frequency</b></td></tr>"
    for emailEntry in emailUpdateDict.keys():

        # Only look at email keys
        if emailEntry.endswith(FREQUENCY):
            continue

        html += "<tr><td>"+ emailEntry +"</td>"
        html += "<td>"+ str( len(emailUpdateDict[emailEntry]) ) +"</td>"
        html += "<td>"+ emailUpdateDict[emailEntry + FREQUENCY] +"</td>"
        html += "</tr>"

    html += "</table>"
    html += "</form>"

    return html

def setEmailUpdates(form):

    pageChoiceList = []

    for pageName in form.keys():

        if not pageName[0].isupper():
            continue

        pageChoiceList.append(pageName)

    emailUpdateDict = {}
    if os.path.exists(emailUpdateFilename):
        emailUpdateDict = load(emailUpdateFilename) 

    emailUpdateDict[email] = pageChoiceList

    frequency = FREQUENCY_DEFAULT
    if form.has_key(FREQUENCY):
        frequency = form[FREQUENCY].value

    emailUpdateDict[email + FREQUENCY] = frequency

    save(emailUpdateFilename, emailUpdateDict)

"""
Sends an e-mail to 'email' saying that the pages in 'pageList' have changed,
and that this email is every 'frequency'.
Use 'server' (an smtplib.SMTP object)
"""
def sendEmail(email, pageList, frequency, server):

    hostname = os.environ.get("HTTP_HOST", "")

    message = "To: "+ email +"\n"
    message += "Subject: Scribble "+ frequency +" update\n"
    message += "Date: "+ email.Utils.formatdate()
    message += "\n"
    message += "The following pages to which you are subscribed have changed "
    message += "since the last "+ frequency +" update\n\n"
    for page in pageList:
        message += page +"\n"
        message += "  Take me there:\n"
        message += "   http://"+ hostname + scriptURL +"?action=view&page="+ page +"\n"
        if USE_RCS:
            message += "  Show what changed:\n"
            message += "   http://"+ hostname + scriptURL +"?action=diff&page="+ page +"\n" 
        message += "\n"

    server.sendmail(SMTP_FROM, email, message)

"""
Build and send e-mail updates
mailFrequency is FREQUENCY_DAILY or FREQUENCY_MONTHLY (see statics at top of file)
"""
def sendEmailUpdates(mailFrequency):

    html = ""

    html += "Frequency: "+ mailFrequency +"<br>"

    if not os.path.exists(emailUpdateFilename):
        html += "File not found: "+ emailUpdateFilename +"<br>"
        return  html

    emailUpdateDict = load(emailUpdateFilename)

    server = smtplib.SMTP(SMTP_SERVER)
    if SMTP_AUTHENTICATE:
        server.login(SMTP_USER, SMTP_PASSWORD)

    html += "Connected to server<br>"

    for email in emailUpdateDict.keys():

        # Only look at email keys
        if email.endswith(FREQUENCY):
            continue

        html += "User: "+ email

        # Does this user want mail this often
        frequency = emailUpdateDict[email + FREQUENCY]
        if not frequency == mailFrequency:
            continue

        html += " wants mail "+ frequency +"<br>"

        userInterestPages = emailUpdateDict[email]

        timeThosePagesChanged = timePagesChanged(userInterestPages)

        html += "User: "+ email +" is interesed in: "+ ",".join(userInterestPages) +"<br>"
   
        mailThesePages = []

        for date in timeThosePagesChanged.keys():

            if isYesterday(date) and mailFrequency == FREQUENCY_DAILY:
                page = timeThosePagesChanged[date] 
                mailThesePages.append( page )
                html += "Relevant: "+ page +"<br>"

            if isThisWeek(date) and mailFrequency == FREQUENCY_WEEKLY:
                page = timeThosePagesChanged[date]
                mailThesePages.append( page )
                html += "Relevant: "+ page +"<br>"

        if len(mailThesePages) > 0:
            sendEmail(email, mailThesePages, mailFrequency, server)

    server.quit()

    return html

# The special keys which can be put in the page as ##KEYNAME## and the methods which we call to get the text of it
# Special keys are all caps and underscores only.
specialKeyToMethodMap = { 
    "COLOR_INPUT_TABLE": colorInputTable,
    "EMAIL_UPDATE_FORM": emailUpdateForm }

#
# MAIN
#

form = cgi.FieldStorage()

if form.has_key("action"):
    action = form["action"].value
else:
    action = "view"

if action != "viewimage":
    print "Content-type: text/html"
    print 

    print "<html>\n"

if form.has_key("page") and isValidFilename( form["page"].value ):
    page = form["page"].value
else:
    page = START_PAGE

email = ""
if form.has_key("email"):
    email = form["email"].value

emailUpdatesPageList = []
filename = fullpath(page)

if action == "view":

    print header(page)
    print bodyStart()

    print view(filename, page)

elif action == "edit":

    print header("Editing "+ page)
    print bodyStart()

    print edit(filename, page)

elif action == "save":

    print header("Saved "+ page)
    print bodyStart()

    if not form.has_key("body"):
        print "<b>**ERROR**</b>: Save action. No 'body' attribute found"
    else:

        if USE_RCS:

            try:
                rcs = rcslib.RCS()
                rcs.checkout(filename, 1, "-f")

            except OSError:
                # This means file not yet in version control. Ignore
                None

        file = open( filename, "wt" )

        contents = form["body"].value
        file.write(contents)

        file.close()

        if USE_RCS:
                rcs = rcslib.RCS()
                rcs.checkin(filename)

        # After a save we go to view
        print view(filename, page) 

elif action == "search":

    print header("Search "+ page)
    print bodyStart()

    if not form.has_key("term"):
        print "<b>**ERROR**</b>: No search term received."
    else:
        term = form["term"].value
        print search(term)

elif action == "whatsnew":
    
    print header("What's New")
    print bodyStart()

    print whatsnew()

elif action == "diff" and USE_RCS:

    print header("Changes")
    print bodyStart()

    if form.has_key("back"):
            back = int( form["back"].value )
    else:
            back = 0

    print rcsdiff(page, back)

elif action == "saveimage":

    print header("Uploaded image")
    print bodyStart()

    if form.has_key("image"):

        try:
            saveimage( form["image"], page )
        
            # Go back to editing the page
            print edit(filename, page)

        except ValueError, e:
            print "<b>Error</b>: Image uploaded failed because:<br>\n"
            print str(e) +"<br>\n"
            print "Click your browsers 'Back' button and try again<br>\n"

elif action == "viewimage":

    print viewimage(filename)

elif action == "colors":

    setColors(form)

    print header(page)
    print bodyStart()

    print view(filename, page)

elif action == "findEmailUpdates":

    if form.has_key("email"):
        emailUpdatesPageList = pageList()

    print header(page)
    print bodyStart()

    print view(filename, page)


elif action == "saveEmailUpdates":

    print header(page)
    print bodyStart()

    setEmailUpdates(form)

    print view(filename, page)

elif action == "sendEmailUpdates":

    frequency = FREQUENCY_DEFAULT
    if form.has_key(FREQUENCY):
        frequency = form[FREQUENCY].value

    print sendEmailUpdates(frequency)

else:
    print header("Error"+ page)
    print bodyStart()

    print "<b>**ERROR**</b>: Invalid action of '"+ action +"' received."


if action != "viewimage":
    print bodyEnd()

